from typing import Optional

from momento_wire_types.controlclient_pb2 import (
    _CreateCacheRequest,
    _CreateSigningKeyRequest,
    _DeleteCacheRequest,
    _ListCachesRequest,
    _ListSigningKeysRequest,
    _RevokeSigningKeyRequest,
)

from .. import _cache_service_errors_converter, logs
from .._utilities._data_validation import _validate_cache_name, _validate_ttl_minutes
from ..cache_operation_types import (
    CreateCacheResponse,
    CreateSigningKeyResponse,
    DeleteCacheResponse,
    ListCachesResponse,
    ListSigningKeysResponse,
    RevokeSigningKeyResponse,
)
from . import _scs_grpc_manager

_DEADLINE_SECONDS = 60.0  # 1 minute


class _ScsControlClient:
    """Momento Internal."""

    def __init__(self, auth_token: str, endpoint: str):
        self._logger = logs.logger
        self._logger.debug("Simple cache control client instantiated with endpoint: %s", endpoint)
        self._grpc_manager = _scs_grpc_manager._ControlGrpcManager(auth_token, endpoint)

    async def create_cache(self, cache_name: str) -> CreateCacheResponse:
        _validate_cache_name(cache_name)
        try:
            self._logger.info(f"Creating cache with name: {cache_name}")
            request = _CreateCacheRequest()
            request.cache_name = cache_name
            await self._grpc_manager.async_stub().CreateCache(request, timeout=_DEADLINE_SECONDS)
            return CreateCacheResponse()
        except Exception as e:
            self._logger.debug("Failed to create cache: %s with exception: %s", cache_name, e)
            raise _cache_service_errors_converter.convert(e) from None

    async def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        _validate_cache_name(cache_name)
        try:
            self._logger.info(f"Deleting cache with name: {cache_name}")
            request = _DeleteCacheRequest()
            request.cache_name = cache_name
            await self._grpc_manager.async_stub().DeleteCache(request, timeout=_DEADLINE_SECONDS)
            return DeleteCacheResponse()
        except Exception as e:
            self._logger.debug("Failed to delete cache: %s with exception: %s", cache_name, e)
            raise _cache_service_errors_converter.convert(e) from None

    async def list_caches(self, next_token: Optional[str] = None) -> ListCachesResponse:
        try:
            list_caches_request = _ListCachesRequest()
            list_caches_request.next_token = next_token if next_token is not None else ""
            return ListCachesResponse.from_grpc_response(
                await self._grpc_manager.async_stub().ListCaches(list_caches_request, timeout=_DEADLINE_SECONDS)
            )
        except Exception as e:
            raise _cache_service_errors_converter.convert(e)

    async def create_signing_key(self, ttl_minutes: int, endpoint: str) -> CreateSigningKeyResponse:
        _validate_ttl_minutes(ttl_minutes)
        try:
            self._logger.info(f"Creating signing key with ttl (in minutes): {ttl_minutes}")
            create_signing_key_request = _CreateSigningKeyRequest()
            create_signing_key_request.ttl_minutes = ttl_minutes
            return CreateSigningKeyResponse.from_grpc_response(
                await self._grpc_manager.async_stub().CreateSigningKey(
                    create_signing_key_request, timeout=_DEADLINE_SECONDS
                ),
                endpoint,
            )
        except Exception as e:
            self._logger.warning(f"Failed to create signing key with exception: {e}")
            raise _cache_service_errors_converter.convert(e)

    async def revoke_signing_key(self, key_id: str) -> RevokeSigningKeyResponse:
        try:
            self._logger.info(f"Revoking signing key with key_id {key_id}")
            request = _RevokeSigningKeyRequest()
            request.key_id = key_id
            await self._grpc_manager.async_stub().RevokeSigningKey(request, timeout=_DEADLINE_SECONDS)
            return RevokeSigningKeyResponse()
        except Exception as e:
            self._logger.warning(f"Failed to revoke signing key with key_id {key_id} exception: {e}")
            raise _cache_service_errors_converter.convert(e)

    async def list_signing_keys(self, endpoint: str, next_token: Optional[str] = None) -> ListSigningKeysResponse:
        try:
            list_signing_keys_request = _ListSigningKeysRequest()
            list_signing_keys_request.next_token = next_token if next_token is not None else ""
            return ListSigningKeysResponse.from_grpc_response(
                await self._grpc_manager.async_stub().ListSigningKeys(
                    list_signing_keys_request, timeout=_DEADLINE_SECONDS
                ),
                endpoint,
            )
        except Exception as e:
            raise _cache_service_errors_converter.convert(e)

    async def close(self) -> None:
        await self._grpc_manager.close()



import json
from enum import Enum
from typing import Dict, Optional
from urllib.parse import quote

import jwt
from jwt.api_jwk import PyJWK

from .errors import InvalidArgumentError


class CacheOperation(Enum):
    GET = 1
    SET = 2


class SigningRequest:
    def __init__(
        self,
        cache_name: str,
        cache_key: str,
        cache_operation: CacheOperation,
        expiry_epoch_seconds: int,
        ttl_seconds: Optional[int] = None,
    ):
        """Initializes a new SigningRequest.

        Args:
            cache_name: The name of the cache.
            cache_key: The key of the object.
            cache_operation: The operation performed on the item in the cache.
            expiry_epoch_seconds: The timestamp that the pre-signed URL is valid until.
            ttl_seconds: Time to Live for the item in Cache. This is an optional property that will only be used for CacheOperation.SET.
        """  # noqa
        self._cache_name = cache_name
        self._cache_key = cache_key
        self._cache_operation = cache_operation
        self._expiry_epoch_seconds = expiry_epoch_seconds
        self._ttl_seconds = ttl_seconds

    def expiry_epoch_seconds(self) -> int:
        return self._expiry_epoch_seconds

    def cache_name(self) -> str:
        return self._cache_name

    def cache_key(self) -> str:
        return self._cache_key

    def cache_operation(self) -> CacheOperation:
        return self._cache_operation

    def ttl_seconds(self) -> Optional[int]:
        return self._ttl_seconds


class MomentoSigner:
    def __init__(self, jwk_json_string: str):
        """Initializes MomentoSigner with the specified private key.

        Args:
            jwk_json_string: the JSON string of the JWK key. This can be obtained from create_signing_key response.

        Raises:
            InvalidArgumentError: If the supplied private key is not valid.
        """
        try:
            self._jwk_json: Dict[str, str] = json.loads(jwk_json_string)
        except json.decoder.JSONDecodeError as e:
            raise InvalidArgumentError(f"Invalid JWK Json String: {jwk_json_string}") from e

        try:
            self._jwk: PyJWK = PyJWK.from_dict(self._jwk_json)  # type: ignore[no-untyped-call, misc]
        except jwt.exceptions.PyJWKError as e:
            raise InvalidArgumentError(f"Invalid JWK: {jwk_json_string}") from e
        except jwt.exceptions.InvalidKeyError as e:
            raise InvalidArgumentError(f"Invalid JWK: {jwk_json_string}") from e

        if self._jwk.key_id is None:  # type: ignore[misc]
            raise InvalidArgumentError(f"JWK missing kid attribute: {jwk_json_string}")

        self._alg = self._jwk_json.get("alg", None)
        if self._alg is None:
            self._alg = self._alg_fallback_logic()

    def sign_access_token(self, signing_request: SigningRequest) -> str:
        """Creates an access token to be used as a JWT token.

        Args:
            signing_request: Contains parameters for the generated token.

        Returns:
            str: the JWT token.
        """
        claims: Dict[str, object] = {
            "exp": signing_request.expiry_epoch_seconds(),
            "cache": signing_request.cache_name(),
            "key": signing_request.cache_key(),
        }

        if signing_request.cache_operation() == CacheOperation.GET:
            claims["method"] = ["get"]
        elif signing_request.cache_operation() == CacheOperation.SET:
            claims["method"] = ["set"]
            if signing_request.ttl_seconds() is None:
                raise InvalidArgumentError("ttl_seconds is required for SET operation.")
            claims["ttl"] = signing_request.ttl_seconds()
        else:
            raise NotImplementedError(f"Unrecognized Operation: {signing_request.cache_operation()}")

        # jwt.encode will automatically insert "typ" and "alg" into the header for us.
        # We still need to specify "kid" to be included in the header however.
        return jwt.encode(
            claims, self._jwk.key, algorithm=self._alg, headers={"kid": self._jwk.key_id}  # type: ignore[misc]
        )

    def create_presigned_url(self, hostname: str, signing_request: SigningRequest) -> str:
        """Creates a pre-signed HTTPS URL.

        Args:
            hostname: Hostname of SimpleCacheService. This value can be obtained from create_signing_key response.
            The SDK will prepend rest to it to get to the REST endpoint
            signing_request: Contains parameters for the generated URL.

        Returns:
            str: a pre-signed HTTPS URL with JWT token.
        """
        token = self.sign_access_token(signing_request)
        cache_name = quote(signing_request.cache_name(), safe="")
        cache_key = quote(signing_request.cache_key(), safe="")
        if signing_request.cache_operation() == CacheOperation.GET:
            return f"https://rest.{hostname}/cache/get/{cache_name}/{cache_key}?token={token}"
        elif signing_request.cache_operation() == CacheOperation.SET:
            ttl_seconds = signing_request.ttl_seconds()
            if ttl_seconds is None:
                raise InvalidArgumentError("ttl_seconds is required for SET operation.")
            url = (
                f"https://rest.{hostname}/cache/set/{cache_name}/{cache_key}"
                f"?token={token}&ttl_milliseconds={ttl_seconds * 1000}"
            )
            return url
        else:
            raise NotImplementedError(f"Unrecognized Operation: {signing_request.cache_operation()}")

    # Logic stolen from https://github.com/jpadilla/pyjwt/blob/master/jwt/api_jwk.py#L19
    # to handle the case when alg is missing from JWK.
    def _alg_fallback_logic(self) -> str:
        kty = self._jwk_json.get("kty", None)
        if kty is None:
            raise InvalidArgumentError("kty is not found: %s" % self._jwk_json)
        crv = self._jwk_json.get("crv", None)
        if kty == "EC":
            if crv == "P-256" or not crv:
                algorithm = "ES256"
            elif crv == "P-384":
                algorithm = "ES384"
            elif crv == "P-521":
                algorithm = "ES512"
            elif crv == "secp256k1":
                algorithm = "ES256K"
            else:
                raise InvalidArgumentError("Unsupported crv: %s" % crv)
        elif kty == "RSA":
            algorithm = "RS256"
        elif kty == "oct":
            algorithm = "HS256"
        elif kty == "OKP":
            if not crv:
                raise InvalidArgumentError("crv is not found: %s" % self._jwk_json)
            if crv == "Ed25519":
                algorithm = "EdDSA"
            else:
                raise InvalidArgumentError("Unsupported crv: %s" % crv)
        else:
            raise InvalidArgumentError("Unsupported kty: %s" % kty)

        return algorithm



from typing import Optional

from momento_wire_types.controlclient_pb2 import (
    _CreateCacheRequest,
    _DeleteCacheRequest,
    _ListCachesRequest,
)
from momento_wire_types.controlclient_pb2_grpc import ScsControlStub

from momento import _cache_service_errors_converter, logs
from momento._utilities._data_validation import _validate_cache_name
from momento.cache_operation_types import (
    CreateCacheResponse,
    DeleteCacheResponse,
    ListCachesResponse,
)

from . import _scs_grpc_manager

_DEADLINE_SECONDS = 60.0  # 1 minute


class _ScsControlClient:
    """Momento Internal."""

    def __init__(self, auth_token: str, endpoint: str):
        self._grpc_manager = _scs_grpc_manager._ControlGrpcManager(auth_token, endpoint)

    def create_cache(self, cache_name: str) -> CreateCacheResponse:
        _validate_cache_name(cache_name)
        try:
            logs.debug(f"Creating cache with name: {cache_name}")
            request = _CreateCacheRequest()
            request.cache_name = cache_name
            self._getStub().CreateCache(request, timeout=_DEADLINE_SECONDS)
            return CreateCacheResponse()
        except Exception as e:
            logs.debug(f"Failed to create cache: {cache_name} with exception:{e}")
            # raise e
            raise _cache_service_errors_converter.convert(e) from None

    def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        _validate_cache_name(cache_name)
        try:
            logs.debug(f"Deleting cache with name: {cache_name}")
            request = _DeleteCacheRequest()
            request.cache_name = cache_name
            self._getStub().DeleteCache(request, timeout=_DEADLINE_SECONDS)
            return DeleteCacheResponse()
        except Exception as e:
            logs.debug(f"Failed to delete cache: {cache_name} with exception:{e}")
            raise _cache_service_errors_converter.convert(e) from None

    def list_caches(self, next_token: Optional[str] = None) -> ListCachesResponse:
        try:
            list_caches_request = _ListCachesRequest()
            list_caches_request.next_token = next_token if next_token is not None else ""
            return ListCachesResponse.from_grpc_response(
                self._getStub().ListCaches(list_caches_request, timeout=_DEADLINE_SECONDS)
            )
        except Exception as e:
            raise _cache_service_errors_converter.convert(e)

    def _getStub(self) -> ScsControlStub:
        return self._grpc_manager.stub()

    def close(self) -> None:
        self._grpc_manager.close()



from types import TracebackType
from typing import Optional, Type, Union

from . import _momento_endpoint_resolver
from ._utilities._data_validation import _validate_request_timeout
from .cache_operation_types import (
    CacheDeleteResponse,
    CacheGetResponse,
    CacheSetResponse,
    CreateCacheResponse,
    DeleteCacheResponse,
    ListCachesResponse,
)
from .internal.synchronous._scs_control_client import _ScsControlClient
from .internal.synchronous._scs_data_client import _ScsDataClient


class SimpleCacheClient:
    def __init__(
        self,
        auth_token: str,
        default_ttl_seconds: int,
        request_timeout_ms: Optional[int] = None,
    ):
        """Creates an async SimpleCacheClient

        Args:
            auth_token (str): Momento Token to authenticate the requests with Simple Cache Service
            default_ttl_seconds (int): A default Time To Live in seconds for cache objects created by this client. It is
                possible to override this setting when calling the set method.
            request_timeout_ms (Optional[int], optional): An optional timeout in milliseconds to allow for Get and Set
                operations to complete. The request will be terminated if it takes longer than this value and will
                result in TimeoutError. Defaults to None, in which case a 5 second timeout is used.
        Raises:
            IllegalArgumentError: If method arguments fail validations.
        """
        _validate_request_timeout(request_timeout_ms)

        endpoints = _momento_endpoint_resolver.resolve(auth_token)
        self._control_client = _ScsControlClient(auth_token, endpoints.control_endpoint)
        self._data_client = _ScsDataClient(
            auth_token,
            endpoints.cache_endpoint,
            default_ttl_seconds,
            request_timeout_ms,
        )

    def __enter__(self) -> "SimpleCacheClient":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._control_client.close()
        self._data_client.close()

    def create_cache(self, cache_name: str) -> CreateCacheResponse:
        """Creates a new cache in your Momento account.

        Args:
            cache_name: String used to create cache.

        Returns:
            CreateCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name is None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            AlreadyExistsError: If cache with the given name already exists.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return self._control_client.create_cache(cache_name)

    def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        """Deletes a cache and all the items within it.

        Args:
            cache_name: String cache name to delete.

        Returns:
            DeleteCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name is None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            NotFoundError: If an attempt is made to delete a MomentoCache that doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return self._control_client.delete_cache(cache_name)

    def list_caches(self, next_token: Optional[str] = None) -> ListCachesResponse:
        """Lists all caches.

        Args:
            next_token: Token to continue paginating through the list. It's used to handle large paginated lists.

        Returns:
            ListCachesResponse

        Raises:
            AuthenticationError: If the provided Momento Auth Token is invalid.
        """
        return self._control_client.list_caches(next_token)

    def set(
        self,
        cache_name: str,
        key: Union[str, bytes],
        value: Union[str, bytes],
        ttl_seconds: Optional[int] = None,
    ) -> CacheSetResponse:
        """Stores an item in cache

        Args:
            cache_name: Name of the cache to store the item in.
            key (string or bytes): The key to be used to store item.
            value (string or bytes): The value to be stored.
            ttl_seconds (Optional): Time to live in cache in seconds. If not provided, then default TTL for the cache
                client instance is used.

        Returns:
            CacheSetResponse

        Raises:
            InvalidArgumentError: If validation fails for the provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to store the item.
        """
        return self._data_client.set(cache_name, key, value, ttl_seconds)

    def get(self, cache_name: str, key: Union[str, bytes]) -> CacheGetResponse:
        """Retrieve an item from the cache

        Args:
            cache_name: Name of the cache to get the item from
            key (string or bytes): The key to be used to retrieve the item.

        Returns:
            CacheGetResponse

        Raises:
            InvalidArgumentError: If validation fails for the provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        return self._data_client.get(cache_name, key)

    def delete(self, cache_name: str, key: Union[str, bytes]) -> CacheDeleteResponse:
        """Delete an item from the cache.

        Performs a no-op if the item is not in the cache.

        Args:
            cache_name: Name of the cache to delete the item from.
            key (string or bytes): The key to delete.

        Returns:
            CacheDeleteResponse

        Raises:
            InvalidArgumentError: If validation fails for provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to delete the item.
        """
        return self._data_client.delete(cache_name, key)



from typing import Optional, Union

from .. import errors

DEFAULT_STRING_CONVERSION_ERROR = "Could not decode bytes to UTF-8"


def _validate_cache_name(cache_name: str) -> None:
    if cache_name is None or not isinstance(cache_name, str):
        raise errors.InvalidArgumentError("Cache name must be a non-empty string")


def _as_bytes(
    data: Union[str, bytes],
    error_message: Optional[str] = DEFAULT_STRING_CONVERSION_ERROR,
) -> bytes:
    if isinstance(data, str):
        return data.encode("utf-8")
    if isinstance(data, bytes):
        return data
    raise errors.InvalidArgumentError(error_message + str(type(data)))


def _validate_ttl(ttl_seconds: int) -> None:
    if not isinstance(ttl_seconds, int) or ttl_seconds < 0:
        raise errors.InvalidArgumentError("TTL Seconds must be a non-negative integer")


def _validate_ttl_minutes(ttl_minutes: int) -> None:
    if not isinstance(ttl_minutes, int) or ttl_minutes < 0:
        raise errors.InvalidArgumentError("TTL Minutes must be a non-negative integer")


def _validate_request_timeout(request_timeout_ms: Optional[int]) -> None:
    if request_timeout_ms is None:
        return
    if not isinstance(request_timeout_ms, int) or request_timeout_ms <= 0:
        raise errors.InvalidArgumentError("Request timeout must be greater than zero.")



from types import TracebackType
from typing import Optional, Type, Union

from .. import logs

try:
    from .._utilities._data_validation import _validate_request_timeout
    from ._scs_control_client import _ScsControlClient
    from ._scs_data_client import _ScsDataClient
except ImportError as e:
    if e.name == "cygrpc":
        import sys

        print(
            "There is an issue on M1 macs between GRPC native packaging and Python wheel tags. "
            "See https://github.com/grpc/grpc/issues/28387",
            file=sys.stderr,
        )
        print("-".join("" for _ in range(99)), file=sys.stderr)
        print("    TO WORK AROUND:", file=sys.stderr)
        print("    * Install Rosetta 2", file=sys.stderr)
        print(
            "    * Install Python from python.org (you might need to do this if you're using an arm-only build)",
            file=sys.stderr,
        )
        print("    * re-run with:", file=sys.stderr)
        print("arch -x86_64 {} {}".format(sys.executable, *sys.argv), file=sys.stderr)
        print("-".join("" for _ in range(99)), file=sys.stderr)
    raise e

from .. import _momento_endpoint_resolver
from ..cache_operation_types import (
    CacheDeleteResponse,
    CacheGetResponse,
    CacheSetResponse,
    CreateCacheResponse,
    CreateSigningKeyResponse,
    DeleteCacheResponse,
    ListCachesResponse,
    ListSigningKeysResponse,
    RevokeSigningKeyResponse,
)


class SimpleCacheClient:
    """Async Simple Cache Client"""

    # For high load, we might get better performance with multiple clients, because the server is
    # configured to allow a max of 100 streams per connection.  In the javascript SDK, multiple
    # clients resulted in an obvious performance improvement.  However, in the python SDK I have
    # not yet been able to observe a clear benefit.  So for now, we are putting the plumbing in
    # place so that we can more easily test performance with multiple connections in the future,
    # but we are leaving the default value set to 1.
    #
    # We are hard-coding the value for now, because we haven't yet designed the API for
    # users to use to configure tunables:
    # https://github.com/momentohq/dev-eco-issue-tracker/issues/85
    _NUM_CLIENTS = 1

    def __init__(
        self,
        auth_token: str,
        default_ttl_seconds: int,
        request_timeout_ms: Optional[int] = None,
    ):
        """Creates an async SimpleCacheClient

        Args:
            auth_token (str): Momento Token to authenticate the requests with Simple Cache Service
            default_ttl_seconds (int): A default Time To Live in seconds for cache objects created by this client. It is
                possible to override this setting when calling the set method.
            request_timeout_ms (Optional[int], optional): An optional timeout in milliseconds to allow for Get and Set
                operations to complete. The request will be terminated if it takes longer than this value and will
                result in TimeoutError. Defaults to None, in which case a 5 second timeout is used.
        Raises:
            IllegalArgumentError: If method arguments fail validations.
        """
        _validate_request_timeout(request_timeout_ms)
        self._logger = logs.logger
        self._next_client_index = 0
        endpoints = _momento_endpoint_resolver.resolve(auth_token)
        self._control_client = _ScsControlClient(auth_token, endpoints.control_endpoint)
        self._data_clients = [
            _ScsDataClient(
                auth_token,
                endpoints.cache_endpoint,
                default_ttl_seconds,
                request_timeout_ms,
            )
            for _ in range(SimpleCacheClient._NUM_CLIENTS)
        ]

    async def __aenter__(self) -> "SimpleCacheClient":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self._control_client.close()
        for data_client in self._data_clients:
            await data_client.close()

    async def create_cache(self, cache_name: str) -> CreateCacheResponse:
        """Creates a new cache in your Momento account.

        Args:
            cache_name: String used to create cache.

        Returns:
            CreateCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            ExistsError: If cache with the given name already exists.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return await self._control_client.create_cache(cache_name)

    async def delete_cache(self, cache_name: str) -> DeleteCacheResponse:
        """Deletes a cache and all of the items within it.

        Args:
            cache_name: String cache name to delete.

        Returns:
            DeleteCacheResponse

        Raises:
            InvalidArgumentError: If provided cache_name is None.
            BadRequestError: If the cache name provided doesn't follow the naming conventions
            NotFoundError: If an attempt is made to delete a MomentoCache that doesn't exits.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return await self._control_client.delete_cache(cache_name)

    async def list_caches(self, next_token: Optional[str] = None) -> ListCachesResponse:
        """Lists all caches.

        Args:
            next_token: Token to continue paginating through the list. It's used to handle large paginated lists.

        Returns:
            ListCachesResponse

        Raises:
            AuthenticationError: If the provided Momento Auth Token is invalid.
        """
        return await self._control_client.list_caches(next_token)

    async def create_signing_key(self, ttl_minutes: int) -> CreateSigningKeyResponse:
        """Creates a Momento signing key

        Args:
            ttl_minutes: The key's time-to-live in minutes

        Returns:
            CreateSigningKeyResponse

        Raises:
            InvalidArgumentError: If provided ttl minutes is negative.
            BadRequestError: If the ttl provided is not accepted
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return await self._control_client.create_signing_key(ttl_minutes, self._get_next_client().get_endpoint())

    async def revoke_signing_key(self, key_id: str) -> RevokeSigningKeyResponse:
        """Revokes a Momento signing key, all tokens signed by which will be invalid

        Args:
            key_id: The id of the Momento signing key to revoke

        Returns:
            RevokeSigningKeyResponse

        Raises:
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return await self._control_client.revoke_signing_key(key_id)

    async def list_signing_keys(self, next_token: Optional[str] = None) -> ListSigningKeysResponse:
        """Lists all Momento signing keys for the provided auth token.

        Args:
            next_token: Token to continue paginating through the list. It's used to handle large paginated lists.

        Returns:
            ListSigningKeysResponse

        Raises:
            AuthenticationError: If the provided Momento Auth Token is invalid.
            ClientSdkError: For any SDK checks that fail.
        """
        return await self._control_client.list_signing_keys(self._get_next_client().get_endpoint(), next_token)

    async def set(
        self,
        cache_name: str,
        key: str,
        value: Union[str, bytes],
        ttl_seconds: Optional[int] = None,
    ) -> CacheSetResponse:
        """Stores an item in cache

        Args:
            cache_name: Name of the cache to store the item in.
            key (string or bytes): The key to be used to store item.
            value (string or bytes): The value to be stored.
            ttl_seconds (Optional): Time to live in cache in seconds. If not provided, then default TTL for the cache
                client instance is used.

        Returns:
            CacheSetResponse

        Raises:
            InvalidArgumentError: If validation fails for provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to store the item.
        """
        return await self._get_next_client().set(cache_name, key, value, ttl_seconds)

    async def get(self, cache_name: str, key: str) -> CacheGetResponse:
        """Retrieve an item from the cache

        Args:
            cache_name: Name of the cache to get the item from
            key (string or bytes): The key to be used to retrieve the item.

        Returns:
            CacheGetResponse

        Raises:
            InvalidArgumentError: If validation fails for provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        return await self._get_next_client().get(cache_name, key)

    async def delete(self, cache_name: str, key: str) -> CacheDeleteResponse:
        """Delete an item from the cache.

        Performs a no-op if the item is not in the cache.

        Args:
            cache_name: Name of the cache to delete the item from.
            key (string or bytes): The key to delete.

        Returns:
            CacheDeleteResponse

        Raises:
            InvalidArgumentError: If validation fails for provided method arguments.
            BadRequestError: If the provided inputs are rejected by server because they are invalid
            NotFoundError: If the cache with the given name doesn't exist.
            AuthenticationError: If the provided Momento Auth Token is invalid.
            InternalServerError: If server encountered an unknown error while trying to delete the item.
        """
        return await self._get_next_client().delete(cache_name, key)

    def _get_next_client(self) -> _ScsDataClient:
        client = self._data_clients[self._next_client_index]
        self._next_client_index = (self._next_client_index + 1) % len(self._data_clients)
        return client



from typing import Optional

import jwt
from jwt.exceptions import DecodeError

from . import errors

_MOMENTO_CONTROL_ENDPOINT_PREFIX = "control."
_MOMENTO_CACHE_ENDPOINT_PREFIX = "cache."
_CONTROL_ENDPOINT_CLAIM_ID = "cp"
_CACHE_ENDPOINT_CLAIM_ID = "c"


class _Endpoints:
    def __init__(self, control_endpoint: str, cache_endpoint: str):
        self.control_endpoint = control_endpoint
        self.cache_endpoint = cache_endpoint


def resolve(auth_token: str, endpoint_override: Optional[str] = None) -> _Endpoints:
    if endpoint_override is not None:
        return _Endpoints(
            _MOMENTO_CONTROL_ENDPOINT_PREFIX + endpoint_override,
            _MOMENTO_CACHE_ENDPOINT_PREFIX + endpoint_override,
        )
    return _getEndpointFromToken(auth_token)


def _getEndpointFromToken(auth_token: str) -> _Endpoints:
    try:
        claims = jwt.decode(auth_token, options={"verify_signature": False})  # type: ignore[misc]
        return _Endpoints(
            claims[_CONTROL_ENDPOINT_CLAIM_ID],  # type: ignore[misc]
            claims[_CACHE_ENDPOINT_CLAIM_ID],  # type: ignore[misc]
        )
    except (DecodeError, KeyError):
        raise errors.InvalidArgumentError("Invalid Auth token.") from None



from typing import Callable, List, Union

import grpc
from grpc.aio import ClientCallDetails, Metadata

from momento.errors import InvalidArgumentError


class Header:
    once_only_headers = ["agent"]

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value


class AddHeaderClientInterceptor(grpc.aio.UnaryUnaryClientInterceptor):
    are_only_once_headers_sent = False

    def __init__(self, headers: List[Header]):
        self._headers_to_add_once: List[Header] = list(
            filter(lambda header: header.name in header.once_only_headers, headers)
        )
        self.headers_to_add_every_time = list(
            filter(lambda header: header.name not in header.once_only_headers, headers)
        )

    async def intercept_unary_unary(
        self,
        continuation: Callable[
            [grpc.aio._interceptor.ClientCallDetails, grpc.aio._typing.RequestType],
            grpc.aio._call.UnaryUnaryCall,
        ],
        client_call_details: grpc.aio._interceptor.ClientCallDetails,
        request: grpc.aio._typing.RequestType,
    ) -> Union[grpc.aio._call.UnaryUnaryCall, grpc.aio._typing.ResponseType]:

        new_client_call_details = sanitize_client_call_details(client_call_details)

        for header in self.headers_to_add_every_time:
            new_client_call_details.metadata.add(header.name, header.value)

        if not AddHeaderClientInterceptor.are_only_once_headers_sent:
            for header in self._headers_to_add_once:
                new_client_call_details.metadata.add(header.name, header.value)
                AddHeaderClientInterceptor.are_only_once_headers_sent = True

        return await continuation(new_client_call_details, request)


def sanitize_client_call_details(client_call_details: grpc.aio.ClientCallDetails) -> grpc.aio.ClientCallDetails:
    """
    Defensive function meant to handle inbound grpc client request objects and make sure we can handle properly
    when we inject our own metadata onto request object. This was mainly done as temporary fix after we observed
    ddtrace grpc client interceptor passing client_call_details.metadata as a list instead of a grpc.aio.Metadata
    object. See this ticket for follow-up actions to come back in and address this longer term:
        https://github.com/momentohq/client-sdk-python/issues/149
    Args:
        client_call_details: the original inbound client grpc request we are intercepting

    Returns: a new client_call_details object with metadata properly initialized to a `grpc.aio.Metadata` object
    """
    new_client_call_details = None
    # If no metadata set on passed in client call details then we are first to set, so we should just initialize
    if client_call_details.metadata is None:
        new_client_call_details = ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=Metadata(),
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )

    # This is block hit when ddtrace interceptor runs first and sets metadata as a list
    elif isinstance(client_call_details.metadata, list):
        existing_headers = client_call_details.metadata
        metadata = Metadata()
        # re-add all existing values to new metadata
        for md_key, md_value in existing_headers:
            metadata.add(md_key, md_value)
        new_client_call_details = ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )
    elif isinstance(client_call_details.metadata, grpc.aio.Metadata):
        # If  proper grpc `grpc.aio.Metadata()` object is passed just use original object passed and pass back
        new_client_call_details = client_call_details
    else:
        # Else we raise exception for now since we don't know how to handle an unknown type
        raise InvalidArgumentError(
            "unexpected grpc client request metadata property passed to interceptor "
            "type=" + str(type(client_call_details.metadata))
        )

    return new_client_call_details



import logging
from typing import Callable, List, Union

import grpc

# TODO: Retry interceptor behavior should be configurable, but we need to
# align on basic API design first:
# https://github.com/momentohq/client-sdk-javascript/issues/79
# https://github.com/momentohq/dev-eco-issue-tracker/issues/85
# For now, for convenience during development, you can toggle this hard-coded
# variable to enable/disable it.
import momento.errors

RETRIES_ENABLED = True
MAX_ATTEMPTS = 3


RETRYABLE_STATUS_CODES: List[grpc.StatusCode] = [
    # # including all the status codes for reference, but
    # # commenting out the ones we don't want to retry on for now.
    # grpc.StatusCode.OK,
    # grpc.StatusCode.CANCELLED,
    # grpc.StatusCode.UNKNOWN,
    # grpc.StatusCode.INVALID_ARGUMENT,
    # grpc.StatusCode.DEADLINE_EXCEEDED,
    # grpc.StatusCode.NOT_FOUND,
    # grpc.StatusCode.ALREADY_EXISTS,
    # grpc.StatusCode.PERMISSION_DENIED,
    # grpc.StatusCode.RESOURCE_EXHAUSTED,
    # grpc.StatusCode.FAILED_PRECONDITION,
    # grpc.StatusCode.ABORTED,
    # grpc.StatusCode.OUT_OF_RANGE,
    # grpc.StatusCode.UNIMPLEMENTED,
    grpc.StatusCode.INTERNAL,
    grpc.StatusCode.UNAVAILABLE,
    # grpc.StatusCode.DATA_LOSS,
    # grpc.StatusCode.UNAUTHENTICATED,
]

LOGGER = logging.getLogger("retry-interceptor")


# TODO: We need to send retry count information to the server so that we
# will have some visibility into how often this is happening to customers:
# https://github.com/momentohq/client-sdk-javascript/issues/80
# TODO: we need to add backoff/jitter for the retries:
# https://github.com/momentohq/client-sdk-javascript/issues/81
def get_retry_interceptor_if_enabled() -> List[grpc.aio.UnaryUnaryClientInterceptor]:
    if not RETRIES_ENABLED:
        return []

    return [RetryInterceptor()]


class RetryInterceptor(grpc.aio.UnaryUnaryClientInterceptor):
    async def intercept_unary_unary(
        self,
        continuation: Callable[
            [grpc.aio._interceptor.ClientCallDetails, grpc.aio._typing.RequestType],
            grpc.aio._call.UnaryUnaryCall,
        ],
        client_call_details: grpc.aio._interceptor.ClientCallDetails,
        request: grpc.aio._typing.RequestType,
    ) -> Union[grpc.aio._call.UnaryUnaryCall, grpc.aio._typing.ResponseType]:
        for try_i in range(MAX_ATTEMPTS):
            call = await continuation(client_call_details, request)
            response_code = await call.code()

            if response_code == grpc.StatusCode.OK:
                return call

            # Return if it was last attempt
            if try_i == (MAX_ATTEMPTS - 1):
                LOGGER.debug(
                    "Request path: %s; retryable status code: %s; number of retries (%i) "
                    "has exceeded max (%i), not retrying.",
                    client_call_details.method.decode("utf-8"),
                    response_code,
                    try_i,
                    MAX_ATTEMPTS,
                )
                return call

            # If status code is not in retryable status codes
            if response_code not in RETRYABLE_STATUS_CODES:
                return call

            LOGGER.debug(
                "Request path: %s; retryable status code: %s; number of retries (%i) "
                "is less than max (%i), retrying.",
                client_call_details.method.decode("utf-8"),
                response_code,
                try_i,
                MAX_ATTEMPTS,
            )

        raise momento.errors.ClientSdkError("Failed to return from RetryInterceptor!  This is a bug.")



from typing import List, Optional, Tuple, Union

from momento_wire_types.cacheclient_pb2 import (
    _DeleteRequest,
    _DeleteResponse,
    _GetRequest,
    _GetResponse,
    _SetRequest,
    _SetResponse,
)
from momento_wire_types.cacheclient_pb2_grpc import ScsStub

from momento import cache_operation_types
from momento._utilities._data_validation import _validate_ttl

from ..common._data_client_ops import (
    construct_delete_response,
    construct_get_response,
    construct_set_response,
    prepare_delete_request,
    prepare_get_request,
    prepare_set_request,
    wrap_with_error_handling,
)
from . import _scs_grpc_manager

_DEFAULT_DEADLINE_SECONDS = 5.0  # 5 seconds


def _make_metadata(cache_name: str) -> List[Tuple[str, str]]:
    return [("cache", cache_name)]


class _ScsDataClient:
    """Internal"""

    def __init__(
        self,
        auth_token: str,
        endpoint: str,
        default_ttl_seconds: int,
        request_timeout_ms: Optional[int],
    ):
        self._default_deadline_seconds = (
            _DEFAULT_DEADLINE_SECONDS if not request_timeout_ms else request_timeout_ms / 1000.0
        )
        self._grpc_manager = _scs_grpc_manager._DataGrpcManager(auth_token, endpoint)
        _validate_ttl(default_ttl_seconds)
        self._default_ttlSeconds = default_ttl_seconds

    def set(
        self,
        cache_name: str,
        key: Union[str, bytes],
        value: Union[str, bytes],
        ttl_seconds: Optional[int],
    ) -> cache_operation_types.CacheSetResponse:
        def execute_set_request_fn(req: _SetRequest) -> _SetResponse:
            return self._getStub().Set(
                req,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )

        return wrap_with_error_handling(
            cache_name=cache_name,
            request_type="Set",
            prepare_request_fn=lambda: prepare_set_request(  # type: ignore[no-any-return]
                key=key, value=value, ttl_seconds=ttl_seconds, default_ttl_seconds=self._default_ttlSeconds
            ),
            execute_request_fn=execute_set_request_fn,
            response_fn=construct_set_response,
        )

    def get(self, cache_name: str, key: Union[str, bytes]) -> cache_operation_types.CacheGetResponse:
        def execute_get_request_fn(req: _GetRequest) -> _GetResponse:
            return self._getStub().Get(
                req,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )

        return wrap_with_error_handling(
            cache_name=cache_name,
            request_type="Get",
            prepare_request_fn=lambda: prepare_get_request(key),  # type: ignore[no-any-return]
            execute_request_fn=execute_get_request_fn,
            response_fn=construct_get_response,
        )

    def delete(self, cache_name: str, key: Union[str, bytes]) -> cache_operation_types.CacheDeleteResponse:
        def execute_delete_request_fn(req: _DeleteRequest) -> _DeleteResponse:
            return self._getStub().Delete(
                req,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )

        return wrap_with_error_handling(
            cache_name=cache_name,
            request_type="Delete",
            prepare_request_fn=lambda: prepare_delete_request(key),  # type: ignore[no-any-return]
            execute_request_fn=execute_delete_request_fn,
            response_fn=construct_delete_response,
        )

    def _getStub(self) -> ScsStub:
        return self._grpc_manager.stub()

    def close(self) -> None:
        self._grpc_manager.close()



import json
from datetime import datetime
from enum import Enum
from typing import Any, List, Optional

from momento_wire_types import cacheclient_pb2 as cache_client_types

from . import _cache_service_errors_converter as error_converter
from . import logs


class CacheGetStatus(Enum):
    HIT = 1
    MISS = 2


class CacheSetResponse:
    def __init__(self, key: bytes, value: bytes):
        """Initializes CacheSetResponse to handle gRPC set response.

        Args:
            key (bytes): The value of the key of item that was stored in cache..
            value (bytes): The value of item that was stored in the cache.
        """
        self._value = value
        self._key = key

    def value(self) -> str:
        """Decodes string value set in cache to a utf-8 string."""
        return self._value.decode("utf-8")

    def value_as_bytes(self) -> bytes:
        """Returns byte value set in cache."""
        return self._value

    def key(self) -> str:
        """Decodes key of item set in cache to a utf-8 string."""
        return self._key.decode("utf-8")

    def key_as_bytes(self) -> bytes:
        """Returns key of item stored in cache as bytes."""
        return self._key

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"CacheSetResponse(key={self._key!r}, value={self._value!r})"


class CacheGetResponse:
    def __init__(self, value: bytes, status: CacheGetStatus):
        self._value = value
        self._status = status

    @staticmethod
    def from_grpc_response(  # type: ignore[misc]
        grpc_get_response: cache_client_types._GetResponse,
    ) -> "CacheGetResponse":
        """Initializes CacheGetResponse to handle gRPC get response.

        Args:
            grpc_get_response: Protobuf based response returned by Scs.

        Raises:
            InternalServerError: If server encountered an unknown error while trying to retrieve the item.
        """
        value: bytes = grpc_get_response.cache_body  # type: ignore[misc]

        if grpc_get_response.result == cache_client_types.Hit:  # type: ignore[misc]
            status = CacheGetStatus.HIT
        elif grpc_get_response.result == cache_client_types.Miss:  # type: ignore[misc]
            status = CacheGetStatus.MISS
        else:
            logs.debug("Get received unsupported ECacheResult: %s", grpc_get_response.result)  # type: ignore[misc]
            raise error_converter.convert_ecache_result(
                grpc_get_response.result, grpc_get_response.message, "GET"  # type: ignore[misc]
            )
        return CacheGetResponse(value=value, status=status)

    def value(self) -> Optional[str]:
        """Returns value stored in cache as utf-8 string if there was Hit. Returns None otherwise."""
        if self._status == CacheGetStatus.HIT:
            return self._value.decode("utf-8")
        return None

    def value_as_bytes(self) -> Optional[bytes]:
        """Returns value stored in cache as bytes if there was Hit. Returns None otherwise."""
        if self._status == CacheGetStatus.HIT:
            return self._value
        return None

    def status(self) -> CacheGetStatus:
        """Returns get operation result such as HIT or MISS."""
        return self._status

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"CacheGetResponse(value={self._value!r}, status={self._status!r})"


class CacheDeleteResponse:
    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "CacheDeleteResponse()"


class CreateCacheResponse:
    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "CreateCacheResponse()"


class DeleteCacheResponse:
    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "DeleteCacheResponse()"


class CacheInfo:
    def __init__(self, name: str):
        """Initializes CacheInfo to handle caches returned from list cache operation.

        Args:
            name (str): Name of the cache.
        """
        self._name = name

    def name(self) -> str:
        """Returns the cache's name."""
        return self._name

    @staticmethod
    def from_grpc_response(grpc_listed_cache: Any) -> "CacheInfo":  # type: ignore[misc]
        """Initializes CacheInfo to handle caches returned from list cache operation.

        Args:
            grpc_listed_cache: Protobuf based response returned by Scs.
        """
        return CacheInfo(name=grpc_listed_cache.cache_name)  # type: ignore[misc]

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"CacheInfo(name={self._name!r})"


class ListCachesResponse:
    def __init__(self, next_token: Optional[str], caches: List[CacheInfo]):
        """Initializes ListCacheResponse to handle list cache response.

        Args:
            next_token (Optional[str]): Next list caches page token.
            caches (List[CacheInfo]): Cache info from this page of results.
        """
        self._next_token = next_token
        self._caches = caches

    def next_token(self) -> Optional[str]:
        """Returns next token."""
        return self._next_token

    def caches(self) -> List[CacheInfo]:
        """Returns all caches."""
        return self._caches

    @staticmethod
    def from_grpc_response(grpc_list_cache_response: Any) -> "ListCachesResponse":  # type: ignore[misc]
        """Initializes ListCacheResponse to handle list cache response.

        Args:
            grpc_list_cache_response: Protobuf based response returned by Scs.
        """
        next_token: Optional[str] = (
            grpc_list_cache_response.next_token  # type: ignore[misc]
            if grpc_list_cache_response.next_token != ""  # type: ignore[misc]
            else None
        )
        caches = [CacheInfo.from_grpc_response(cache) for cache in grpc_list_cache_response.cache]  # type: ignore[misc]
        return ListCachesResponse(next_token=next_token, caches=caches)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"ListCachesResponse(next_token={self._next_token!r}, caches={self._caches!r})"


class CreateSigningKeyResponse:
    def __init__(self, key_id: str, endpoint: str, key: str, expires_at: datetime):
        """Initializes CreateSigningKeyResponse to handle create signing key response.

        Args:
            grpc_create_signing_key_response: Protobuf based response returned by Scs.
        """
        self._key_id = key_id
        self._endpoint = endpoint
        self._key = key
        self._expires_at = expires_at

    def key_id(self) -> str:
        """Returns the id of the signing key"""
        return self._key_id

    def endpoint(self) -> str:
        """Returns the endpoint of the signing key"""
        return self._endpoint

    def key(self) -> str:
        """Returns the JSON string of the key itself"""
        return self._key

    def expires_at(self) -> datetime:
        """Returns the datetime representation of when the key expires"""
        return self._expires_at

    @staticmethod
    def from_grpc_response(  # type: ignore[misc]
        grpc_create_signing_key_response: Any, endpoint: str
    ) -> "CreateSigningKeyResponse":
        key_id: str = json.loads(grpc_create_signing_key_response.key)["kid"]  # type: ignore[misc]
        key: str = grpc_create_signing_key_response.key  # type: ignore[misc]
        expires_at: datetime = datetime.fromtimestamp(grpc_create_signing_key_response.expires_at)  # type: ignore[misc]
        return CreateSigningKeyResponse(key_id, endpoint, key, expires_at)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return (
            f"CreateSigningKeyResponse(key_id={self._key_id!r}, endpoint={self._endpoint!r}, "
            f"key={self._key!r}, expires_at={self._expires_at!r})"
        )


class RevokeSigningKeyResponse:
    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "RevokeSigningKeyResponse()"


class SigningKey:
    def __init__(self, key_id: str, expires_at: datetime, endpoint: str):
        """Initializes SigningKey to handle signing keys returned from list signing keys operation.

        Args:
            grpc_listed_signing_key: Protobuf based response returned by Scs.
        """
        self._key_id = key_id
        self._expires_at = expires_at
        self._endpoint = endpoint

    def key_id(self) -> str:
        """Returns the id of the Momento signing key"""
        return self._key_id

    def expires_at(self) -> datetime:
        """Returns the time the key expires"""
        return self._expires_at

    def endpoint(self) -> str:
        """Returns the endpoint of the Momento signing key"""
        return self._endpoint

    @staticmethod
    def from_grpc_response(grpc_listed_signing_key: Any, endpoint: str) -> "SigningKey":  # type: ignore[misc]
        key_id: str = grpc_listed_signing_key.key_id  # type: ignore[misc]
        expires_at: datetime = datetime.fromtimestamp(grpc_listed_signing_key.expires_at)  # type: ignore[misc]
        return SigningKey(key_id, expires_at, endpoint)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"SigningKey(key_id={self._key_id!r}, expires_at={self._expires_at!r}, endpoint={self._endpoint!r})"


class ListSigningKeysResponse:
    def __init__(self, next_token: Optional[str], signing_keys: List[SigningKey]):
        """Initializes ListSigningKeysResponse to handle list signing keys response.

        Args:
            grpc_list_signing_keys_response: Protobuf based response returned by Scs.
        """
        self._next_token = next_token
        self._signing_keys = signing_keys

    def next_token(self) -> Optional[str]:
        """Returns next token."""
        return self._next_token

    def signing_keys(self) -> List[SigningKey]:
        """Returns all signing keys."""
        return self._signing_keys

    @staticmethod
    def from_grpc_response(  # type: ignore[misc]
        grpc_list_signing_keys_response: Any, endpoint: str
    ) -> "ListSigningKeysResponse":
        next_token: Optional[str] = (
            grpc_list_signing_keys_response.next_token  # type: ignore[misc]
            if grpc_list_signing_keys_response.next_token != ""  # type: ignore[misc]
            else None
        )
        signing_keys: List[SigningKey] = [
            SigningKey.from_grpc_response(signing_key, endpoint)  # type: ignore[misc]
            for signing_key in grpc_list_signing_keys_response.signing_key  # type: ignore[misc]
        ]
        return ListSigningKeysResponse(next_token, signing_keys)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"ListSigningKeysResponse(next_token={self._next_token!r}, signing_keys={self._signing_keys!r})"



from typing import Awaitable, Callable, Optional, TypeVar, Union

from momento_wire_types.cacheclient_pb2 import (
    _DeleteRequest,
    _DeleteResponse,
    _GetRequest,
    _GetResponse,
    _SetRequest,
    _SetResponse,
)

from momento import _cache_service_errors_converter, cache_operation_types, logs
from momento._utilities._data_validation import (
    _as_bytes,
    _validate_cache_name,
    _validate_ttl,
)

TResponse = TypeVar("TResponse")
TGeneratedRequest = TypeVar("TGeneratedRequest")
TGeneratedResponse = TypeVar("TGeneratedResponse")
TExecuteResult = TypeVar("TExecuteResult")
TMomentoResponse = TypeVar("TMomentoResponse")

_logger = logs.logger


def wrap_with_error_handling(
    cache_name: str,
    request_type: str,
    prepare_request_fn: Callable[[], TGeneratedRequest],
    execute_request_fn: Callable[[TGeneratedRequest], TGeneratedResponse],
    response_fn: Callable[[TGeneratedRequest, TGeneratedResponse], TMomentoResponse],
) -> TMomentoResponse:
    _validate_cache_name(cache_name)
    try:
        req = prepare_request_fn()
        resp = execute_request_fn(req)
        return response_fn(req, resp)
    except Exception as e:
        _logger.warning("%s failed with exception: %s", request_type, e)
        raise _cache_service_errors_converter.convert(e)


async def wrap_async_with_error_handling(
    cache_name: str,
    request_type: str,
    prepare_request_fn: Callable[[], TGeneratedRequest],
    execute_request_fn: Callable[[TGeneratedRequest], Awaitable[TGeneratedResponse]],
    response_fn: Callable[[TGeneratedRequest, TGeneratedResponse], TMomentoResponse],
) -> TMomentoResponse:
    _validate_cache_name(cache_name)
    try:
        req = prepare_request_fn()
        resp = await execute_request_fn(req)
        return response_fn(req, resp)
    except Exception as e:
        _logger.warning("%s failed with exception: %s", request_type, e)
        raise _cache_service_errors_converter.convert(e)


def prepare_set_request(
    key: Union[str, bytes],
    value: Union[str, bytes],
    ttl_seconds: Optional[int],
    default_ttl_seconds: int,
) -> _GetRequest:
    _logger.log(logs.TRACE, "Issuing a set request with key %s", str(key))
    item_ttl_seconds = default_ttl_seconds if ttl_seconds is None else ttl_seconds
    _validate_ttl(item_ttl_seconds)
    set_request = _SetRequest()
    set_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
    set_request.cache_body = _as_bytes(value, "Unsupported type for value: ")
    set_request.ttl_milliseconds = item_ttl_seconds * 1000
    return set_request


def construct_set_response(req: _SetRequest, resp: _SetResponse) -> cache_operation_types.CacheSetResponse:
    _logger.log(logs.TRACE, "Set succeeded for key: %s", str(req.cache_key))
    return cache_operation_types.CacheSetResponse(req.cache_key, req.cache_body)


def prepare_get_request(key: Union[str, bytes]) -> _GetRequest:
    _logger.log(logs.TRACE, "Issuing a Get request with key %s", str(key))
    get_request = _GetRequest()
    get_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
    return get_request


def construct_get_response(req: _GetRequest, resp: _GetResponse) -> cache_operation_types.CacheGetResponse:
    _logger.log(logs.TRACE, "Received a get response for %s", str(req.cache_key))
    return cache_operation_types.CacheGetResponse.from_grpc_response(resp)


def prepare_delete_request(key: Union[str, bytes]) -> _DeleteRequest:
    _logger.log(logs.TRACE, "Issuing a Delete request with key %s", str(key))
    delete_request = _DeleteRequest()
    delete_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
    return delete_request


def construct_delete_response(req: _DeleteRequest, resp: _DeleteResponse) -> cache_operation_types.CacheDeleteResponse:
    _logger.log(logs.TRACE, "Received a delete response for %s", str(req.cache_key))
    return cache_operation_types.CacheDeleteResponse()



import collections
from typing import Callable, List, TypeVar

import grpc

from momento.errors import InvalidArgumentError

RequestType = TypeVar("RequestType")
ResponseType = TypeVar("ResponseType")


class Header:
    once_only_headers = ["agent"]

    def __init__(self, name: str, value: str):
        self.name = name
        self.value = value


class _ClientCallDetails(
    collections.namedtuple("_ClientCallDetails", ("method", "timeout", "metadata", "credentials")),
    grpc.ClientCallDetails,
):
    pass


class AddHeaderClientInterceptor(grpc.UnaryUnaryClientInterceptor):
    are_only_once_headers_sent = False

    @staticmethod
    def is_only_once_header(header: Header) -> bool:
        return header.name in header.once_only_headers

    @staticmethod
    def is_not_only_once_header(header: Header) -> bool:
        return header.name not in header.once_only_headers

    def __init__(self, headers: List[Header]):
        self._headers_to_add_once: List[Header] = list(filter(AddHeaderClientInterceptor.is_only_once_header, headers))
        self.headers_to_add_every_time = list(filter(AddHeaderClientInterceptor.is_not_only_once_header, headers))

    def intercept_unary_unary(
        self,
        continuation: Callable[[grpc.ClientCallDetails, RequestType], grpc.Call],
        client_call_details: grpc.ClientCallDetails,
        request: RequestType,
    ) -> grpc.Call:
        new_client_call_details = sanitize_client_call_details(client_call_details)

        for header in self.headers_to_add_every_time:
            new_client_call_details.metadata.append((header.name, header.value))

        if not AddHeaderClientInterceptor.are_only_once_headers_sent:
            for header in self._headers_to_add_once:
                new_client_call_details.metadata.append((header.name, header.value))
                AddHeaderClientInterceptor.are_only_once_headers_sent = True

        return continuation(new_client_call_details, request)


def sanitize_client_call_details(client_call_details: grpc.ClientCallDetails) -> grpc.ClientCallDetails:
    """
    Defensive function meant to handle inbound grpc client request objects and make sure we can handle properly
    when we inject our own metadata onto request object. This was mainly done as temporary fix after we observed
    ddtrace grpc client interceptor passing client_call_details.metadata as a list instead of a grpc.aio.Metadata
    object. See this ticket for follow-up actions to come back in and address this longer term:
        https://github.com/momentohq/client-sdk-python/issues/149

    In the synchronous gRPC implementation, grpc.ClientCallDetails (vs grpc.aio.ClientCallDetails) does actually
    use a list of tuples to represent the metadata.  So we do that here, which is different from our behavior on
    the asynchronous client.

    Args:
        client_call_details: the original inbound client grpc request we are intercepting

    Returns: a new client_call_details object with metadata properly initialized
    """

    # If no metadata set on passed in client call details then we are first to set, so we should just initialize
    if client_call_details.metadata is None:
        return _ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=[],
            credentials=client_call_details.credentials,
        )

    # This is block hit when ddtrace interceptor runs first and sets metadata as a list
    elif isinstance(client_call_details.metadata, list):
        return client_call_details
    else:
        # Else we raise exception for now since we don't know how to handle an unknown type
        raise InvalidArgumentError(
            "unexpected grpc client request metadata property passed to interceptor "
            "type=" + str(type(client_call_details.metadata))
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['momento',
 'momento._utilities',
 'momento.aio',
 'momento.internal',
 'momento.internal.common',
 'momento.internal.synchronous']

package_data = \
{'': ['*']}

install_requires = \
['cryptography==3.4.8',
 'grpcio==1.50.0',
 'momento-wire-types==0.31.1',
 'pyjwt==2.4.0']

setup_kwargs = {
    'name': 'momento',
    'version': '0.18.0',
    'description': 'SDK for Momento',
    'long_description': '# Momento client-sdk-python\n\n:warning: Experimental SDK :warning:\n\nPython SDK for Momento is experimental and under active development.\nThere could be non-backward compatible changes or removal in the future.\nPlease be aware that you may need to update your source code with the current version of the SDK when its version gets upgraded.\n\n---\n\n<br/>\n\nPython SDK for Momento, a serverless cache that automatically scales without any of the operational overhead required by traditional caching solutions.\n\n<br/>\n\n## Getting Started :running:\n\n### Requirements\n\n- [Python 3.7](https://www.python.org/downloads/) or above is required\n- A Momento Auth Token is required, you can generate one using the [Momento CLI](https://github.com/momentohq/momento-cli)\n\n<br/>\n\n### Installing Momento and Running the Example\n\nCheck out our [Python SDK example](/examples/)!\n\n<br/>\n\n### Using Momento\n\n```python\nimport os\nfrom momento import simple_cache_client as scc\n\n# Initializing Momento\n_MOMENTO_AUTH_TOKEN = os.getenv(\'MOMENTO_AUTH_TOKEN\')\n_ITEM_DEFAULT_TTL_SECONDS = 60\nwith scc.SimpleCacheClient(_MOMENTO_AUTH_TOKEN, _ITEM_DEFAULT_TTL_SECONDS) as cache_client:\n    # Creating a cache named "cache"\n    _CACHE_NAME = \'cache\'\n    cache_client.create_cache(_CACHE_NAME)\n\n    # Sets key with default TTL and get value with that key\n    _KEY = \'MyKey\'\n    _VALUE = \'MyValue\'\n    cache_client.set(_CACHE_NAME, _KEY, _VALUE)\n    get_resp = cache_client.get(_CACHE_NAME, _KEY)\n    print(f\'Looked up Value: {str(get_resp.value())}\')\n\n    # Sets key with TTL of 5 seconds\n    cache_client.set(_CACHE_NAME, _KEY, _VALUE, 5)\n\n    # Permanently deletes cache\n    cache_client.delete_cache(_CACHE_NAME)\n```\n\n<br/>\n',
    'author': 'Momento',
    'author_email': 'hello@momentohq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gomomento.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.12',
}


setup(**setup_kwargs)



from typing import List

import grpc
import momento_wire_types.cacheclient_pb2_grpc as cache_client
import momento_wire_types.controlclient_pb2_grpc as control_client
import pkg_resources

from momento.internal.synchronous._add_header_client_interceptor import (
    AddHeaderClientInterceptor,
    Header,
)


class _ControlGrpcManager:
    """Momento Internal."""

    version = pkg_resources.get_distribution("momento").version

    def __init__(self, auth_token: str, endpoint: str) -> None:
        self._secure_channel = grpc.secure_channel(endpoint, grpc.ssl_channel_credentials())
        intercept_channel = grpc.intercept_channel(self._secure_channel, *_interceptors(auth_token))
        self._stub = control_client.ScsControlStub(intercept_channel)

    def close(self) -> None:
        self._secure_channel.close()

    def stub(self) -> control_client.ScsControlStub:
        return self._stub


class _DataGrpcManager:
    """Momento Internal."""

    version = pkg_resources.get_distribution("momento").version

    def __init__(self, auth_token: str, endpoint: str):
        self._secure_channel = grpc.secure_channel(
            target=endpoint,
            credentials=grpc.ssl_channel_credentials(),
        )
        intercept_channel = grpc.intercept_channel(self._secure_channel, *_interceptors(auth_token))
        self._stub = cache_client.ScsStub(intercept_channel)

    def close(self) -> None:
        self._secure_channel.close()

    def stub(self) -> cache_client.ScsStub:
        return self._stub


def _interceptors(auth_token: str) -> List[grpc.UnaryUnaryClientInterceptor]:
    headers = [Header("authorization", auth_token), Header("agent", f"python:{_ControlGrpcManager.version}")]
    return [AddHeaderClientInterceptor(headers)]



from typing import Any

import grpc

from . import errors, logs

__rpc_to_error = {
    grpc.StatusCode.INVALID_ARGUMENT: errors.BadRequestError,
    grpc.StatusCode.OUT_OF_RANGE: errors.BadRequestError,
    grpc.StatusCode.UNIMPLEMENTED: errors.BadRequestError,
    grpc.StatusCode.FAILED_PRECONDITION: errors.BadRequestError,
    grpc.StatusCode.CANCELLED: errors.CancelledError,
    grpc.StatusCode.DEADLINE_EXCEEDED: errors.TimeoutError,
    grpc.StatusCode.PERMISSION_DENIED: errors.PermissionError,
    grpc.StatusCode.UNAUTHENTICATED: errors.AuthenticationError,
    grpc.StatusCode.RESOURCE_EXHAUSTED: errors.LimitExceededError,
    grpc.StatusCode.ALREADY_EXISTS: errors.AlreadyExistsError,
    grpc.StatusCode.NOT_FOUND: errors.NotFoundError,
    grpc.StatusCode.UNKNOWN: errors.InternalServerError,
    grpc.StatusCode.ABORTED: errors.InternalServerError,
    grpc.StatusCode.INTERNAL: errors.InternalServerError,
    grpc.StatusCode.UNAVAILABLE: errors.InternalServerError,
    grpc.StatusCode.DATA_LOSS: errors.InternalServerError,
}


def convert(exception: Exception) -> Exception:
    if isinstance(exception, errors.SdkError):
        return exception

    if isinstance(exception, grpc.RpcError):
        if exception.code() in __rpc_to_error:
            return __rpc_to_error[exception.code()](exception.details())
        else:
            return errors.InternalServerError("CacheService failed with an internal error")

    return errors.ClientSdkError("Operation failed with error: " + str(exception))


def convert_ecache_result(  # type: ignore[misc]
    ecache_result: Any, message: str, operation_name: str
) -> errors.InternalServerError:
    logs.debug("Converting ECacheResult: %s to error.", ecache_result)
    return errors.InternalServerError(
        f"CacheService returned an unexpected result: {ecache_result}"
        + f" for operation: {operation_name} with message: {message}"
    )
